/***************************************************************************
 * Copyright 2013 Kieker Project (http://kieker-monitoring.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/
package kieker.webgui.persistence.impl.utility;

import java.io.IOException;
import java.lang.reflect.Modifier;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.jar.JarEntry;
import java.util.jar.JarInputStream;

import javax.annotation.Nullable;

import com.google.common.base.Predicate;
import com.google.common.collect.Collections2;

import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;
import kieker.webgui.common.ClassAndMethodContainer;

import org.springframework.stereotype.Service;

/**
 * This tool class can be used to find all plugins and repositories within a given jar file - assume that the given class loader knows these jars.
 * 
 * @author Nils Christian Ehmke
 */
@Service
public class PluginFinder {

	private static final Log LOG = LogFactory.getLog(PluginFinder.class);

	/**
	 * Default constructor. <b>Do not use this constructor. This bean is Spring managed.</b>
	 */
	public PluginFinder() {
		// No code necessary
	}

	/**
	 * This method delivers all classes which are available in the given jar and are compatible with {@link kieker.analysis.repository.AbstractRepository}. None of
	 * them is abstract.
	 * 
	 * @param url
	 *            The URL for the jar.
	 * @param classLoader
	 *            The class loader which should be used to load the classes.
	 * @param classAndMethodContainer
	 *            The container for the necessary reflection methods.
	 * 
	 * @return A list containing all available repository-classes or null, if an exception occurred.
	 * 
	 * @throws IOException
	 *             If something went wrong during the opening of the jar file.
	 */
	public Collection<Class<?>> getAllRepositoriesWithinJar(final URL url, final ClassLoader classLoader, final ClassAndMethodContainer classAndMethodContainer)
			throws IOException {
		// Get all classes within the Jar. Filter the classes which are not repositories. Filter the abstract classes.
		return Collections2.filter(Collections2.filter(this.getAllClassesWithinJar(url, classLoader), new IsRepositoryPredicate(classAndMethodContainer)),
				new IsNotAbstractPredicate());
	}

	/**
	 * This method delivers all classes which are available in the given jar and have the Plugin-Annotation from the Kieker framework (And are correctly compatible
	 * with {@link kieker.analysis.plugin.reader.AbstractReaderPlugin}). None of them is abstract.
	 * 
	 * @param url
	 *            The URL for the jar.
	 * @param classLoader
	 *            The class loader which should be used to load the classes.
	 * @param classAndMethodContainer
	 *            The container for the necessary reflection methods.
	 * 
	 * @return A list containing all available plugin-classes or null, if an exception occurred.
	 * 
	 * @throws IOException
	 *             If something went wrong during the opening of the jar file.
	 */
	public Collection<Class<?>> getAllReadersWithinJar(final URL url, final ClassLoader classLoader, final ClassAndMethodContainer classAndMethodContainer)
			throws IOException {
		// Get all classes within the Jar. Filter the classes which are not readers. Filter the abstract classes.
		return Collections2.filter(Collections2.filter(this.getAllClassesWithinJar(url, classLoader), new IsReaderPredicate(classAndMethodContainer)),
				new IsNotAbstractPredicate());
	}

	/**
	 * This method delivers all classes which are available in the given jar and have the Plugin-Annotation from the Kieker framework (And are correctly compatible
	 * with {@link kieker.analysis.plugin.filter.AbstractFilterPlugin}). None of them is abstract.
	 * 
	 * @param url
	 *            The URL for the jar.
	 * @param classLoader
	 *            The class loader which should be used to load the classes.
	 * @param classAndMethodContainer
	 *            The container for the necessary reflection methods.
	 * 
	 * @return A list containing all available plugin-classes or null, if an exception occurred.
	 * 
	 * @throws IOException
	 *             If something went wrong during the opening of the jar file.
	 */
	public Collection<Class<?>> getAllFiltersWithinJar(final URL url, final ClassLoader classLoader, final ClassAndMethodContainer classAndMethodContainer)
			throws IOException {
		// Get all classes within the Jar. Filter the classes which are not filters. Filter the abstract classes.
		return Collections2.filter(Collections2.filter(this.getAllClassesWithinJar(url, classLoader), new IsFilterPredicate(classAndMethodContainer)),
				new IsNotAbstractPredicate());
	}

	/**
	 * This method delivers all classes which are available in the given jar.
	 * 
	 * @param url
	 *            The URL for the jar.
	 * @param classLoader
	 *            The class loader which should be used to load the classes.
	 * 
	 * @return A list containing all available classes or null, if an exception occurred.
	 * 
	 * @throws IOException
	 *             If something went wrong during the opening of the jar file.
	 */
	private Collection<Class<?>> getAllClassesWithinJar(final URL url, final ClassLoader classLoader) throws IOException {
		JarInputStream stream = null;
		try {
			final List<Class<?>> result = new ArrayList<Class<?>>();
			// Open the jar as a stream and run through all entries within this file
			stream = new JarInputStream(url.openStream());

			JarEntry jarEntry;
			while ((jarEntry = stream.getNextJarEntry()) != null) {
				try {
					// Assemble the correct name
					final String className = jarEntry.toString().replace('/', '.').replace(".class", "");

					// Try to find a class with the same name and put it into our list
					final Class<?> c = classLoader.loadClass(className);
					result.add(c);
				} catch (final Throwable ex) { // NOPMD (Generic throwable and empty catch block) NOCS (IllegalCatchCheck)
					// Ignore error.
				}
			}
			// Don't forget to close the jar file again.
			stream.close();
			return result;
		} finally {
			if (stream != null) {
				try {
					stream.close();
				} catch (final IOException ex1) {
					// Log but ignore this fail.
					LOG.warn("Could not close stream.", ex1);
				}
			}
		}
	}

	/**
	 * A predicate which can be used to check whether a class is a repository or not.
	 * 
	 * @author Nils Christian Ehmke
	 */
	private static class IsRepositoryPredicate implements Predicate<Class<?>> {

		private final ClassAndMethodContainer classAndMethodContainer;

		public IsRepositoryPredicate(final ClassAndMethodContainer classAndMethodContainer) {
			this.classAndMethodContainer = classAndMethodContainer;
		}

		@Override
		public boolean apply(@Nullable final Class<?> elem) {
			return (elem != null) && elem.isAnnotationPresent(this.classAndMethodContainer.getRepositoryAnnotationClass())
					&& this.classAndMethodContainer.getAbstractRepositoryClass().isAssignableFrom(elem);
		}

	}

	/**
	 * A predicate which can be used to check whether a class is a reader or not.
	 * 
	 * @author Nils Christian Ehmke
	 */
	private static class IsReaderPredicate implements Predicate<Class<?>> {

		private final ClassAndMethodContainer classAndMethodContainer;

		public IsReaderPredicate(final ClassAndMethodContainer classAndMethodContainer) {
			this.classAndMethodContainer = classAndMethodContainer;
		}

		@Override
		public boolean apply(@Nullable final Class<?> elem) {
			return (elem != null) && elem.isAnnotationPresent(this.classAndMethodContainer.getPluginAnnotationClass())
					&& this.classAndMethodContainer.getAbstractReaderPluginClass().isAssignableFrom(elem);
		}

	}

	/**
	 * A predicate which can be used to check whether a class is a filter or not.
	 * 
	 * @author Nils Christian Ehmke
	 */
	private static class IsFilterPredicate implements Predicate<Class<?>> {

		private final ClassAndMethodContainer classAndMethodContainer;

		public IsFilterPredicate(final ClassAndMethodContainer classAndMethodContainer) {
			this.classAndMethodContainer = classAndMethodContainer;
		}

		@Override
		public boolean apply(@Nullable final Class<?> elem) {
			return (elem != null) && elem.isAnnotationPresent(this.classAndMethodContainer.getPluginAnnotationClass())
					&& this.classAndMethodContainer.getAbstractFilterPluginClass().isAssignableFrom(elem);
		}

	}

	/**
	 * A predicate which can be used to check whether a class is abstract or not.
	 * 
	 * @author Nils Christian Ehmke
	 */
	private static class IsNotAbstractPredicate implements Predicate<Class<?>> {

		public IsNotAbstractPredicate() {
			// No code necessary
		}

		@Override
		public boolean apply(@Nullable final Class<?> elem) {
			return (elem != null) && !Modifier.isAbstract(elem.getModifiers());
		}

	}

}
